#include "td_main.h"
#include <raymath.h>
#include <stdlib.h>
#include <math.h>

//# Variables
GUIState guiState = {0};
GameTime gameTime = {0};
Tower towers[TOWER_MAX_COUNT];
int towerCount = 0;

Level levels[] = {
  [0] = {
    .state = LEVEL_STATE_BUILDING,
    .initialGold = 10,
    .waves[0] = {
      .enemyType = ENEMY_TYPE_MINION,
      .wave = 0,
      .count = 10,
      .interval = 1.0f,
      .delay = 1.0f,
      .spawnPosition = {0, 6},
    },
    .waves[1] = {
      .enemyType = ENEMY_TYPE_MINION,
      .wave = 1,
      .count = 20,
      .interval = 0.5f,
      .delay = 1.0f,
      .spawnPosition = {0, 6},
    },
    .waves[2] = {
      .enemyType = ENEMY_TYPE_MINION,
      .wave = 2,
      .count = 30,
      .interval = 0.25f,
      .delay = 1.0f,
      .spawnPosition = {0, 6},
    }
  },
};

Level *currentLevel = levels;

Particle particles[PARTICLE_MAX_COUNT];
int particleCount = 0;

//# Particle system

void ParticleInit()
{
  for (int i = 0; i < PARTICLE_MAX_COUNT; i++)
  {
    particles[i] = (Particle){0};
  }
  particleCount = 0;
}

void ParticleAdd(uint8_t particleType, Vector3 position, Vector3 velocity, float lifetime)
{
  if (particleCount >= PARTICLE_MAX_COUNT)
  {
    return;
  }

  int index = -1;
  for (int i = 0; i < particleCount; i++)
  {
    if (particles[i].particleType == PARTICLE_TYPE_NONE)
    {
      index = i;
      break;
    }
  }

  if (index == -1)
  {
    index = particleCount++;
  }

  Particle *particle = &particles[index];
  particle->particleType = particleType;
  particle->spawnTime = gameTime.time;
  particle->lifetime = lifetime;
  particle->position = position;
  particle->velocity = velocity;
}

void ParticleUpdate()
{
  for (int i = 0; i < particleCount; i++)
  {
    Particle *particle = &particles[i];
    if (particle->particleType == PARTICLE_TYPE_NONE)
    {
      continue;
    }

    float age = gameTime.time - particle->spawnTime;

    if (particle->lifetime > age)
    {
      particle->position = Vector3Add(particle->position, Vector3Scale(particle->velocity, gameTime.deltaTime));
    }
    else {
      particle->particleType = PARTICLE_TYPE_NONE;
    }
  }
}

void DrawExplosionParticle(Particle *particle, float transition)
{
  float size = 1.2f * (1.0f - transition);
  Color startColor = WHITE;
  Color endColor = RED;
  Color color = ColorLerp(startColor, endColor, transition);
  DrawCube(particle->position, size, size, size, color);
}

void ParticleDraw()
{
  for (int i = 0; i < particleCount; i++)
  {
    Particle particle = particles[i];
    if (particle.particleType == PARTICLE_TYPE_NONE)
    {
      continue;
    }

    float age = gameTime.time - particle.spawnTime;
    float transition = age / particle.lifetime;
    switch (particle.particleType)
    {
    case PARTICLE_TYPE_EXPLOSION:
      DrawExplosionParticle(&particle, transition);
      break;
    default:
      DrawCube(particle.position, 0.3f, 0.5f, 0.3f, RED);
      break;
    }
  }
}

//# Pathfinding map

// The queue is a simple array of nodes, we add nodes to the end and remove
// nodes from the front. We keep the array around to avoid unnecessary allocations
static PathfindingNode *pathfindingNodeQueue = 0;
static int pathfindingNodeQueueCount = 0;
static int pathfindingNodeQueueCapacity = 0;

// The pathfinding map stores the distances from the castle to each cell in the map.
PathfindingMap pathfindingMap = {0};

void PathfindingMapInit(int width, int height, Vector3 translate, float scale)
{
  // transforming between map space and world space allows us to adapt 
  // position and scale of the map without changing the pathfinding data
  pathfindingMap.toWorldSpace = MatrixTranslate(translate.x, translate.y, translate.z);
  pathfindingMap.toWorldSpace = MatrixMultiply(pathfindingMap.toWorldSpace, MatrixScale(scale, scale, scale));
  pathfindingMap.toMapSpace = MatrixInvert(pathfindingMap.toWorldSpace);
  pathfindingMap.width = width;
  pathfindingMap.height = height;
  pathfindingMap.scale = scale;
  pathfindingMap.distances = (float *)MemAlloc(width * height * sizeof(float));
  for (int i = 0; i < width * height; i++)
  {
    pathfindingMap.distances[i] = -1.0f;
  }

  pathfindingMap.towerIndex = (long *)MemAlloc(width * height * sizeof(long));
  pathfindingMap.deltaSrc = (DeltaSrc *)MemAlloc(width * height * sizeof(DeltaSrc));
}

float PathFindingGetDistance(int mapX, int mapY)
{
  if (mapX < 0 || mapX >= pathfindingMap.width || mapY < 0 || mapY >= pathfindingMap.height)
  {
    // when outside the map, we return the manhattan distance to the castle (0,0)
    return fabsf((float)mapX) + fabsf((float)mapY);
  }

  return pathfindingMap.distances[mapY * pathfindingMap.width + mapX];
}

void PathFindingNodePush(int16_t x, int16_t y, int16_t fromX, int16_t fromY, float distance)
{
  if (pathfindingNodeQueueCount >= pathfindingNodeQueueCapacity)
  {
    pathfindingNodeQueueCapacity = pathfindingNodeQueueCapacity == 0 ? 256 : pathfindingNodeQueueCapacity * 2;
    // we use MemAlloc/MemRealloc to allocate memory for the queue
    // I am not entirely sure if MemRealloc allows passing a null pointer
    // so we check if the pointer is null and use MemAlloc in that case
    if (pathfindingNodeQueue == 0)
    {
      pathfindingNodeQueue = (PathfindingNode *)MemAlloc(pathfindingNodeQueueCapacity * sizeof(PathfindingNode));
    }
    else
    {
      pathfindingNodeQueue = (PathfindingNode *)MemRealloc(pathfindingNodeQueue, pathfindingNodeQueueCapacity * sizeof(PathfindingNode));
    }
  }

  PathfindingNode *node = &pathfindingNodeQueue[pathfindingNodeQueueCount++];
  node->x = x;
  node->y = y;
  node->fromX = fromX;
  node->fromY = fromY;
  node->distance = distance;
}

PathfindingNode *PathFindingNodePop()
{
  if (pathfindingNodeQueueCount == 0)
  {
    return 0;
  }
  // we return the first node in the queue; we want to return a pointer to the node
  // so we can return 0 if the queue is empty. 
  // We should _not_ return a pointer to the element in the list, because the list
  // may be reallocated and the pointer would become invalid. Or the 
  // popped element is overwritten by the next push operation.
  // Using static here means that the variable is permanently allocated.
  static PathfindingNode node;
  node = pathfindingNodeQueue[0];
  // we shift all nodes one position to the front
  for (int i = 1; i < pathfindingNodeQueueCount; i++)
  {
    pathfindingNodeQueue[i - 1] = pathfindingNodeQueue[i];
  }
  --pathfindingNodeQueueCount;
  return &node;
}

// transform a world position to a map position in the array; 
// returns true if the position is inside the map
int PathFindingFromWorldToMapPosition(Vector3 worldPosition, int16_t *mapX, int16_t *mapY)
{
  Vector3 mapPosition = Vector3Transform(worldPosition, pathfindingMap.toMapSpace);
  *mapX = (int16_t)mapPosition.x;
  *mapY = (int16_t)mapPosition.z;
  return *mapX >= 0 && *mapX < pathfindingMap.width && *mapY >= 0 && *mapY < pathfindingMap.height;
}

void PathFindingMapUpdate()
{
  const int castleX = 0, castleY = 0;
  int16_t castleMapX, castleMapY;
  if (!PathFindingFromWorldToMapPosition((Vector3){castleX, 0.0f, castleY}, &castleMapX, &castleMapY))
  {
    return;
  }
  int width = pathfindingMap.width, height = pathfindingMap.height;

  // reset the distances to -1
  for (int i = 0; i < width * height; i++)
  {
    pathfindingMap.distances[i] = -1.0f;
  }
  // reset the tower indices
  for (int i = 0; i < width * height; i++)
  {
    pathfindingMap.towerIndex[i] = -1;
  }
  // reset the delta src
  for (int i = 0; i < width * height; i++)
  {
    pathfindingMap.deltaSrc[i].x = 0;
    pathfindingMap.deltaSrc[i].y = 0;
  }

  for (int i = 0; i < towerCount; i++)
  {
    Tower *tower = &towers[i];
    if (tower->towerType == TOWER_TYPE_NONE || tower->towerType == TOWER_TYPE_BASE)
    {
      continue;
    }
    int16_t mapX, mapY;
    // technically, if the tower cell scale is not in sync with the pathfinding map scale,
    // this would not work correctly and needs to be refined to allow towers covering multiple cells
    // or having multiple towers in one cell; for simplicity, we assume that the tower covers exactly
    // one cell. For now.
    if (!PathFindingFromWorldToMapPosition((Vector3){tower->x, 0.0f, tower->y}, &mapX, &mapY))
    {
      continue;
    }
    int index = mapY * width + mapX;
    pathfindingMap.towerIndex[index] = i;
  }

  // we start at the castle and add the castle to the queue
  pathfindingMap.maxDistance = 0.0f;
  pathfindingNodeQueueCount = 0;
  PathFindingNodePush(castleMapX, castleMapY, castleMapX, castleMapY, 0.0f);
  PathfindingNode *node = 0;
  while ((node = PathFindingNodePop()))
  {
    if (node->x < 0 || node->x >= width || node->y < 0 || node->y >= height)
    {
      continue;
    }
    int index = node->y * width + node->x;
    if (pathfindingMap.distances[index] >= 0 && pathfindingMap.distances[index] <= node->distance)
    {
      continue;
    }

    int deltaX = node->x - node->fromX;
    int deltaY = node->y - node->fromY;
    // even if the cell is blocked by a tower, we still may want to store the direction
    // (though this might not be needed, IDK right now)
    pathfindingMap.deltaSrc[index].x = (char) deltaX;
    pathfindingMap.deltaSrc[index].y = (char) deltaY;

    // we skip nodes that are blocked by towers
    if (pathfindingMap.towerIndex[index] >= 0)
    {
      node->distance += 8.0f;
    }
    pathfindingMap.distances[index] = node->distance;
    pathfindingMap.maxDistance = fmaxf(pathfindingMap.maxDistance, node->distance);
    PathFindingNodePush(node->x, node->y + 1, node->x, node->y, node->distance + 1.0f);
    PathFindingNodePush(node->x, node->y - 1, node->x, node->y, node->distance + 1.0f);
    PathFindingNodePush(node->x + 1, node->y, node->x, node->y, node->distance + 1.0f);
    PathFindingNodePush(node->x - 1, node->y, node->x, node->y, node->distance + 1.0f);
  }
}

void PathFindingMapDraw()
{
  float cellSize = pathfindingMap.scale * 0.9f;
  float highlightDistance = fmodf(GetTime() * 4.0f, pathfindingMap.maxDistance);
  for (int x = 0; x < pathfindingMap.width; x++)
  {
    for (int y = 0; y < pathfindingMap.height; y++)
    {
      float distance = pathfindingMap.distances[y * pathfindingMap.width + x];
      float colorV = distance < 0 ? 0 : fminf(distance / pathfindingMap.maxDistance, 1.0f);
      Color color = distance < 0 ? BLUE : (Color){fminf(colorV, 1.0f) * 255, 0, 0, 255};
      Vector3 position = Vector3Transform((Vector3){x, -0.25f, y}, pathfindingMap.toWorldSpace);
      // animate the distance "wave" to show how the pathfinding algorithm expands
      // from the castle
      if (distance + 0.5f > highlightDistance && distance - 0.5f < highlightDistance)
      {
        color = BLACK;
      }
      DrawCube(position, cellSize, 0.1f, cellSize, color);
    }
  }
}

Vector2 PathFindingGetGradient(Vector3 world)
{
  int16_t mapX, mapY;
  if (PathFindingFromWorldToMapPosition(world, &mapX, &mapY))
  {
    DeltaSrc delta = pathfindingMap.deltaSrc[mapY * pathfindingMap.width + mapX];
    return (Vector2){(float)-delta.x, (float)-delta.y};
  }
  // fallback to a simple gradient calculation
  float n = PathFindingGetDistance(mapX, mapY - 1);
  float s = PathFindingGetDistance(mapX, mapY + 1);
  float w = PathFindingGetDistance(mapX - 1, mapY);
  float e = PathFindingGetDistance(mapX + 1, mapY);
  return (Vector2){w - e + 0.25f, n - s + 0.125f};
}

//# Projectiles
Projectile projectiles[PROJECTILE_MAX_COUNT];
int projectileCount = 0;

void ProjectileInit()
{
  for (int i = 0; i < PROJECTILE_MAX_COUNT; i++)
  {
    projectiles[i] = (Projectile){0};
  }
}

void ProjectileDraw()
{
  for (int i = 0; i < projectileCount; i++)
  {
    Projectile projectile = projectiles[i];
    if (projectile.projectileType == PROJECTILE_TYPE_NONE)
    {
      continue;
    }
    float transition = (gameTime.time - projectile.shootTime) / (projectile.arrivalTime - projectile.shootTime);
    if (transition >= 1.0f)
    {
      continue;
    }
    Vector2 position = Vector2Lerp(projectile.position, projectile.target, transition);
    float x = position.x;
    float y = position.y;
    float dx = projectile.directionNormal.x;
    float dy = projectile.directionNormal.y;
    for (float d = 1.0f; d > 0.0f; d -= 0.25f)
    {
      x -= dx * 0.1f;
      y -= dy * 0.1f;
      float size = 0.1f * d;
      DrawCube((Vector3){x, 0.2f, y}, size, size, size, RED);
    }
  }
}

void ProjectileUpdate()
{
  for (int i = 0; i < projectileCount; i++)
  {
    Projectile *projectile = &projectiles[i];
    if (projectile->projectileType == PROJECTILE_TYPE_NONE)
    {
      continue;
    }
    float transition = (gameTime.time - projectile->shootTime) / (projectile->arrivalTime - projectile->shootTime);
    if (transition >= 1.0f)
    {
      projectile->projectileType = PROJECTILE_TYPE_NONE;
      Enemy *enemy = EnemyTryResolve(projectile->targetEnemy);
      if (enemy)
      {
        EnemyAddDamage(enemy, projectile->damage);
      }
      continue;
    }
  }
}

Projectile *ProjectileTryAdd(uint8_t projectileType, Enemy *enemy, Vector2 position, Vector2 target, float speed, float damage)
{
  for (int i = 0; i < PROJECTILE_MAX_COUNT; i++)
  {
    Projectile *projectile = &projectiles[i];
    if (projectile->projectileType == PROJECTILE_TYPE_NONE)
    {
      projectile->projectileType = projectileType;
      projectile->shootTime = gameTime.time;
      projectile->arrivalTime = gameTime.time + Vector2Distance(position, target) / speed;
      projectile->damage = damage;
      projectile->position = position;
      projectile->target = target;
      projectile->directionNormal = Vector2Normalize(Vector2Subtract(target, position));
      projectile->targetEnemy = EnemyGetId(enemy);
      projectileCount = projectileCount <= i ? i + 1 : projectileCount;
      return projectile;
    }
  }
  return 0;
}

//# Towers

void TowerInit()
{
  for (int i = 0; i < TOWER_MAX_COUNT; i++)
  {
    towers[i] = (Tower){0};
  }
  towerCount = 0;
}

Tower *TowerGetAt(int16_t x, int16_t y)
{
  for (int i = 0; i < towerCount; i++)
  {
    if (towers[i].x == x && towers[i].y == y)
    {
      return &towers[i];
    }
  }
  return 0;
}

Tower *TowerTryAdd(uint8_t towerType, int16_t x, int16_t y)
{
  if (towerCount >= TOWER_MAX_COUNT)
  {
    return 0;
  }

  Tower *tower = TowerGetAt(x, y);
  if (tower)
  {
    return 0;
  }

  tower = &towers[towerCount++];
  tower->x = x;
  tower->y = y;
  tower->towerType = towerType;
  tower->cooldown = 0.0f;
  tower->damage = 0.0f;
  return tower;
}

Tower *GetTowerByType(uint8_t towerType)
{
  for (int i = 0; i < towerCount; i++)
  {
    if (towers[i].towerType == towerType)
    {
      return &towers[i];
    }
  }
  return 0;
}

int GetTowerCosts(uint8_t towerType)
{
  switch (towerType)
  {
  case TOWER_TYPE_BASE:
    return 0;
  case TOWER_TYPE_GUN:
    return 6;
  case TOWER_TYPE_WALL:
    return 2;
  }
  return 0;
}

float TowerGetMaxHealth(Tower *tower)
{
  switch (tower->towerType)
  {
  case TOWER_TYPE_BASE:
    return 10.0f;
  case TOWER_TYPE_GUN:
    return 3.0f;
  case TOWER_TYPE_WALL:
    return 5.0f;
  }
  return 0.0f;
}

void TowerDraw()
{
  for (int i = 0; i < towerCount; i++)
  {
    Tower tower = towers[i];
    DrawCube((Vector3){tower.x, 0.125f, tower.y}, 1.0f, 0.25f, 1.0f, GRAY);
    switch (tower.towerType)
    {
    case TOWER_TYPE_BASE:
      DrawCube((Vector3){tower.x, 0.4f, tower.y}, 0.8f, 0.8f, 0.8f, MAROON);
      break;
    case TOWER_TYPE_GUN:
      DrawCube((Vector3){tower.x, 0.2f, tower.y}, 0.8f, 0.4f, 0.8f, DARKPURPLE);
      break;
    case TOWER_TYPE_WALL:
      DrawCube((Vector3){tower.x, 0.5f, tower.y}, 1.0f, 1.0f, 1.0f, LIGHTGRAY);
      break;
    }
  }
}

void TowerGunUpdate(Tower *tower)
{
  if (tower->cooldown <= 0)
  {
    Enemy *enemy = EnemyGetClosestToCastle(tower->x, tower->y, 3.0f);
    if (enemy)
    {
      tower->cooldown = 0.125f;
      // shoot the enemy; determine future position of the enemy
      float bulletSpeed = 1.0f;
      float bulletDamage = 3.0f;
      Vector2 velocity = enemy->simVelocity;
      Vector2 futurePosition = EnemyGetPosition(enemy, gameTime.time - enemy->startMovingTime, &velocity, 0);
      Vector2 towerPosition = {tower->x, tower->y};
      float eta = Vector2Distance(towerPosition, futurePosition) / bulletSpeed;
      for (int i = 0; i < 8; i++) {
        velocity = enemy->simVelocity;
        futurePosition = EnemyGetPosition(enemy, gameTime.time - enemy->startMovingTime + eta, &velocity, 0);
        float distance = Vector2Distance(towerPosition, futurePosition);
        float eta2 = distance / bulletSpeed;
        if (fabs(eta - eta2) < 0.01f) {
          break;
        }
        eta = (eta2 + eta) * 0.5f;
      }
      ProjectileTryAdd(PROJECTILE_TYPE_BULLET, enemy, towerPosition, futurePosition, 
        bulletSpeed, bulletDamage);
      enemy->futureDamage += bulletDamage;
    }
  }
  else
  {
    tower->cooldown -= gameTime.deltaTime;
  }
}

void TowerUpdate()
{
  for (int i = 0; i < towerCount; i++)
  {
    Tower *tower = &towers[i];
    switch (tower->towerType)
    {
    case TOWER_TYPE_GUN:
      TowerGunUpdate(tower);
      break;
    }
  }
}

//# Game

void InitLevel(Level *level)
{
  TowerInit();
  EnemyInit();
  ProjectileInit();
  ParticleInit();
  TowerTryAdd(TOWER_TYPE_BASE, 0, 0);

  level->placementMode = 0;
  level->state = LEVEL_STATE_BUILDING;
  level->nextState = LEVEL_STATE_NONE;
  level->playerGold = level->initialGold;

  Camera *camera = &level->camera;
  camera->position = (Vector3){1.0f, 12.0f, 6.5f};
  camera->target = (Vector3){0.0f, 0.5f, 1.0f};
  camera->up = (Vector3){0.0f, 1.0f, 0.0f};
  camera->fovy = 45.0f;
  camera->projection = CAMERA_PERSPECTIVE;
}

void DrawLevelHud(Level *level)
{
  const char *text = TextFormat("Gold: %d", level->playerGold);
  Font font = GetFontDefault();
  DrawTextEx(font, text, (Vector2){GetScreenWidth() - 120, 10}, font.baseSize * 2.0f, 2.0f, BLACK);
  DrawTextEx(font, text, (Vector2){GetScreenWidth() - 122, 8}, font.baseSize * 2.0f, 2.0f, YELLOW);
}

void DrawLevelReportLostWave(Level *level)
{
  BeginMode3D(level->camera);
  DrawGrid(10, 1.0f);
  TowerDraw();
  EnemyDraw();
  ProjectileDraw();
  ParticleDraw();
  guiState.isBlocked = 0;
  EndMode3D();

  const char *text = "Wave lost";
  int textWidth = MeasureText(text, 20);
  DrawText(text, (GetScreenWidth() - textWidth) * 0.5f, 20, 20, WHITE);

  if (Button("Reset level", 20, GetScreenHeight() - 40, 160, 30, 0))
  {
    level->nextState = LEVEL_STATE_RESET;
  }
}

int HasLevelNextWave(Level *level)
{
  for (int i = 0; i < 10; i++)
  {
    EnemyWave *wave = &level->waves[i];
    if (wave->wave == level->currentWave)
    {
      return 1;
    }
  }
  return 0;
}

void DrawLevelReportWonWave(Level *level)
{
  BeginMode3D(level->camera);
  DrawGrid(10, 1.0f);
  TowerDraw();
  EnemyDraw();
  ProjectileDraw();
  ParticleDraw();
  guiState.isBlocked = 0;
  EndMode3D();

  const char *text = "Wave won";
  int textWidth = MeasureText(text, 20);
  DrawText(text, (GetScreenWidth() - textWidth) * 0.5f, 20, 20, WHITE);


  if (Button("Reset level", 20, GetScreenHeight() - 40, 160, 30, 0))
  {
    level->nextState = LEVEL_STATE_RESET;
  }

  if (HasLevelNextWave(level))
  {
    if (Button("Prepare for next wave", GetScreenWidth() - 300, GetScreenHeight() - 40, 300, 30, 0))
    {
      level->nextState = LEVEL_STATE_BUILDING;
    }
  }
  else {
    if (Button("Level won", GetScreenWidth() - 300, GetScreenHeight() - 40, 300, 30, 0))
    {
      level->nextState = LEVEL_STATE_WON_LEVEL;
    }
  }
}

void DrawBuildingBuildButton(Level *level, int x, int y, int width, int height, uint8_t towerType, const char *name)
{
  static ButtonState buttonStates[8] = {0};
  int cost = GetTowerCosts(towerType);
  const char *text = TextFormat("%s: %d", name, cost);
  buttonStates[towerType].isSelected = level->placementMode == towerType;
  buttonStates[towerType].isDisabled = level->playerGold < cost;
  if (Button(text, x, y, width, height, &buttonStates[towerType]))
  {
    level->placementMode = buttonStates[towerType].isSelected ? 0 : towerType;
  }
}

void DrawLevelBuildingState(Level *level)
{
  BeginMode3D(level->camera);
  DrawGrid(10, 1.0f);
  TowerDraw();
  EnemyDraw();
  ProjectileDraw();
  ParticleDraw();

  Ray ray = GetScreenToWorldRay(GetMousePosition(), level->camera);
  float planeDistance = ray.position.y / -ray.direction.y;
  float planeX = ray.direction.x * planeDistance + ray.position.x;
  float planeY = ray.direction.z * planeDistance + ray.position.z;
  int16_t mapX = (int16_t)floorf(planeX + 0.5f);
  int16_t mapY = (int16_t)floorf(planeY + 0.5f);
  if (level->placementMode && !guiState.isBlocked)
  {
    DrawCubeWires((Vector3){mapX, 0.2f, mapY}, 1.0f, 0.4f, 1.0f, RED);
    if (IsMouseButtonPressed(MOUSE_LEFT_BUTTON))
    {
      if (TowerTryAdd(level->placementMode, mapX, mapY))
      {
        level->playerGold -= GetTowerCosts(level->placementMode);
        level->placementMode = TOWER_TYPE_NONE;
      }
    }
  }

  guiState.isBlocked = 0;

  EndMode3D();

  static ButtonState buildWallButtonState = {0};
  static ButtonState buildGunButtonState = {0};
  buildWallButtonState.isSelected = level->placementMode == TOWER_TYPE_WALL;
  buildGunButtonState.isSelected = level->placementMode == TOWER_TYPE_GUN;

  DrawBuildingBuildButton(level, 10, 10, 80, 30, TOWER_TYPE_WALL, "Wall");
  DrawBuildingBuildButton(level, 10, 50, 80, 30, TOWER_TYPE_GUN, "Gun");

  if (Button("Reset level", 20, GetScreenHeight() - 40, 160, 30, 0))
  {
    level->nextState = LEVEL_STATE_RESET;
  }
  
  if (Button("Begin waves", GetScreenWidth() - 160, GetScreenHeight() - 40, 160, 30, 0))
  {
    level->nextState = LEVEL_STATE_BATTLE;
  }

  const char *text = "Building phase";
  int textWidth = MeasureText(text, 20);
  DrawText(text, (GetScreenWidth() - textWidth) * 0.5f, 20, 20, WHITE);
}

void InitBattleStateConditions(Level *level)
{
  level->state = LEVEL_STATE_BATTLE;
  level->nextState = LEVEL_STATE_NONE;
  level->waveEndTimer = 0.0f;
  for (int i = 0; i < 10; i++)
  {
    EnemyWave *wave = &level->waves[i];
    wave->spawned = 0;
    wave->timeToSpawnNext = wave->delay;
  }
}

void DrawLevelBattleState(Level *level)
{
  BeginMode3D(level->camera);
  DrawGrid(10, 1.0f);
  TowerDraw();
  EnemyDraw();
  ProjectileDraw();
  ParticleDraw();
  guiState.isBlocked = 0;
  EndMode3D();

  if (Button("Reset level", 20, GetScreenHeight() - 40, 160, 30, 0))
  {
    level->nextState = LEVEL_STATE_RESET;
  }

  int maxCount = 0;
  int remainingCount = 0;
  for (int i = 0; i < 10; i++)
  {
    EnemyWave *wave = &level->waves[i];
    if (wave->wave != level->currentWave)
    {
      continue;
    }
    maxCount += wave->count;
    remainingCount += wave->count - wave->spawned;
  }
  int aliveCount = EnemyCount();
  remainingCount += aliveCount;

  const char *text = TextFormat("Battle phase: %03d%%", 100 - remainingCount * 100 / maxCount);
  int textWidth = MeasureText(text, 20);
  DrawText(text, (GetScreenWidth() - textWidth) * 0.5f, 20, 20, WHITE);
}

void DrawLevel(Level *level)
{
  switch (level->state)
  {
    case LEVEL_STATE_BUILDING: DrawLevelBuildingState(level); break;
    case LEVEL_STATE_BATTLE: DrawLevelBattleState(level); break;
    case LEVEL_STATE_WON_WAVE: DrawLevelReportWonWave(level); break;
    case LEVEL_STATE_LOST_WAVE: DrawLevelReportLostWave(level); break;
    default: break;
  }

  DrawLevelHud(level);
}

void UpdateLevel(Level *level)
{
  if (level->state == LEVEL_STATE_BATTLE)
  {
    int activeWaves = 0;
    for (int i = 0; i < 10; i++)
    {
      EnemyWave *wave = &level->waves[i];
      if (wave->spawned >= wave->count || wave->wave != level->currentWave)
      {
        continue;
      }
      activeWaves++;
      wave->timeToSpawnNext -= gameTime.deltaTime;
      if (wave->timeToSpawnNext <= 0.0f)
      {
        Enemy *enemy = EnemyTryAdd(wave->enemyType, wave->spawnPosition.x, wave->spawnPosition.y);
        if (enemy)
        {
          wave->timeToSpawnNext = wave->interval;
          wave->spawned++;
        }
      }
    }
    if (GetTowerByType(TOWER_TYPE_BASE) == 0) {
      level->waveEndTimer += gameTime.deltaTime;
      if (level->waveEndTimer >= 2.0f)
      {
        level->nextState = LEVEL_STATE_LOST_WAVE;
      }
    }
    else if (activeWaves == 0 && EnemyCount() == 0)
    {
      level->waveEndTimer += gameTime.deltaTime;
      if (level->waveEndTimer >= 2.0f)
      {
        level->nextState = LEVEL_STATE_WON_WAVE;
      }
    }
  }

  PathFindingMapUpdate();
  EnemyUpdate();
  TowerUpdate();
  ProjectileUpdate();
  ParticleUpdate();

  if (level->nextState == LEVEL_STATE_RESET)
  {
    InitLevel(level);
  }
  
  if (level->nextState == LEVEL_STATE_BATTLE)
  {
    InitBattleStateConditions(level);
  }
  
  if (level->nextState == LEVEL_STATE_WON_WAVE)
  {
    level->currentWave++;
    level->state = LEVEL_STATE_WON_WAVE;
  }
  
  if (level->nextState == LEVEL_STATE_LOST_WAVE)
  {
    level->state = LEVEL_STATE_LOST_WAVE;
  }

  if (level->nextState == LEVEL_STATE_BUILDING)
  {
    level->state = LEVEL_STATE_BUILDING;
  }

  if (level->nextState == LEVEL_STATE_WON_LEVEL)
  {
    // make something of this later
    InitLevel(level);
  }

  level->nextState = LEVEL_STATE_NONE;
}

float nextSpawnTime = 0.0f;

void ResetGame()
{
  InitLevel(currentLevel);
}

void InitGame()
{
  TowerInit();
  EnemyInit();
  ProjectileInit();
  ParticleInit();
  PathfindingMapInit(20, 20, (Vector3){-10.0f, 0.0f, -10.0f}, 1.0f);

  currentLevel = levels;
  InitLevel(currentLevel);
}

//# Immediate GUI functions

int Button(const char *text, int x, int y, int width, int height, ButtonState *state)
{
  Rectangle bounds = {x, y, width, height};
  int isPressed = 0;
  int isSelected = state && state->isSelected;
  if (CheckCollisionPointRec(GetMousePosition(), bounds) && !guiState.isBlocked && !state->isDisabled)
  {
    Color color = isSelected ? DARKGRAY : GRAY;
    DrawRectangle(x, y, width, height, color);
    if (IsMouseButtonPressed(MOUSE_LEFT_BUTTON))
    {
      isPressed = 1;
    }
    guiState.isBlocked = 1;
  }
  else
  {
    Color color = isSelected ? WHITE : LIGHTGRAY;
    DrawRectangle(x, y, width, height, color);
  }
  Font font = GetFontDefault();
  Vector2 textSize = MeasureTextEx(font, text, font.baseSize * 2.0f, 1);
  Color textColor = state->isDisabled ? GRAY : BLACK;
  DrawTextEx(font, text, (Vector2){x + width / 2 - textSize.x / 2, y + height / 2 - textSize.y / 2}, font.baseSize * 2.0f, 1, textColor);
  return isPressed;
}

//# Main game loop

void GameUpdate()
{
  float dt = GetFrameTime();
  // cap maximum delta time to 0.1 seconds to prevent large time steps
  if (dt > 0.1f) dt = 0.1f;
  gameTime.time += dt;
  gameTime.deltaTime = dt;

  UpdateLevel(currentLevel);
}

int main(void)
{
  int screenWidth, screenHeight;
  GetPreferredSize(&screenWidth, &screenHeight);
  InitWindow(screenWidth, screenHeight, "Tower defense");
  SetTargetFPS(30);

  InitGame();

  while (!WindowShouldClose())
  {
    if (IsPaused()) {
      // canvas is not visible in browser - do nothing
      continue;
    }

    BeginDrawing();
    ClearBackground(DARKBLUE);

    GameUpdate();
    DrawLevel(currentLevel);

    EndDrawing();
  }

  CloseWindow();

  return 0;
}